/*
	PluginManager v4
	Copyright (C) 2011, plum

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// header
#include "plugin_mgr.h"

// global
DrawBuffer draw_buf;

// extern
extern u8 msx[];

// function
static __inline u16 convert_8888_to_565(u32 color)
{
	int r, g, b;

	b = (color >> 19) & 0x1F;
	g = (color >> 10) & 0x3F;
	r = (color >>  3) & 0x1F;

	return r | (g << 5) | (b << 11);
}

static __inline u16 convert_8888_to_5551(u32 color)
{
	int r, g, b, a;

	a = (color >> 24) ? 0x8000 : 0;
	b = (color >> 19) & 0x1F;
	g = (color >> 11) & 0x1F;
	r = (color >>  3) & 0x1F;

	return a | r | (g << 5) | (b << 10);
}

static __inline u16 convert_8888_to_4444(u32 color)
{
	int r, g, b, a;

	a = (color >> 28) & 0xF;
	b = (color >> 20) & 0xF;
	g = (color >> 12) & 0xF;
	r = (color >>  4) & 0xF;

	return (a << 12) | r | (g << 4) | (g << 8);
}

static __inline u16 convert_u32_to_u16(u32 color, int pixelformat)
{
	switch(pixelformat)
	{
		case PSP_DISPLAY_PIXEL_FORMAT_565:		return convert_8888_to_565(color);
		case PSP_DISPLAY_PIXEL_FORMAT_5551:		return convert_8888_to_5551(color);
		case PSP_DISPLAY_PIXEL_FORMAT_4444:		return convert_8888_to_4444(color);
	}

	return convert_8888_to_565(color);
}

void GetDrawInfo(void)
{
	int unk;
	sceDisplayGetMode(&unk, &draw_buf.width, &draw_buf.height);
	sceDisplayGetFrameBuf(&draw_buf.vram, &draw_buf.bufferwidth, &draw_buf.pixelformat, PSP_DISPLAY_SETBUF_IMMEDIATE);
	draw_buf.vram = (void *)((u32)draw_buf.vram | 0x44000000);
}

void ClearDisplay(void)
{
	int pixel_size = (draw_buf.pixelformat == PSP_DISPLAY_PIXEL_FORMAT_8888) ? 4 : 2;
	memset(draw_buf.vram, 0, draw_buf.bufferwidth * draw_buf.height * pixel_size);
}

int BlitPrint(int x, int y, u32 fg_col, u32 bg_col, const char *str)
{
	u8 font;
	u32 *vram32, *vram32_ptr;
	u16 *vram16, *vram16_ptr;
	u16 fg16, bg16;
	int i, n, code;
	char *ptr;

	// set
	x <<= 3;
	y <<= 3;
	ptr = (char *)str;

	// draw
	if(draw_buf.pixelformat == PSP_DISPLAY_PIXEL_FORMAT_8888)
	{
		for(; *ptr != '\0' && x < draw_buf.width; ptr++, x += 8)
		{
			code = (*ptr & 0x7F) << 3;
			vram32 = (u32 *)draw_buf.vram + x + y * draw_buf.bufferwidth;

			for(i = 0; i < 8; i++, vram32 += draw_buf.bufferwidth)
			{
				font = msx[code + i];
				vram32_ptr = vram32;

				for(n = 0; n < 8; n++)
				{
					*vram32_ptr++ = (font & 0x80) ? fg_col : bg_col;
					font <<= 1;
				}
			}
		}
	}
	else
	{
		fg16 = convert_u32_to_u16(fg_col, draw_buf.pixelformat);
		bg16 = convert_u32_to_u16(bg_col, draw_buf.pixelformat);

		for(; *ptr != '\0' && x < draw_buf.width; ptr++, x += 8)
		{
			code = (*ptr & 0x7F) << 3;
			vram16 = (u16 *)draw_buf.vram + x + y * draw_buf.bufferwidth;

			for(i = 0; i < 8; i++, vram16 += draw_buf.bufferwidth)
			{
				font = msx[code + i];
				vram16_ptr = vram16;

				for(n = 0; n < 8; n++)
				{
					*vram16_ptr++ = (font & 0x80) ? fg16 : bg16;
					font <<= 1;
				}
			}
		}
	}

	return ptr - str;
}

int BlitPrintf(int x, int y, u32 fg_col, u32 bg_col, const char *format, ...)
{
	va_list ap;
	char str[256];

	va_start(ap, format);
	vsnprintf(str, 256, format, ap);
	va_end(ap);

	return BlitPrint(x, y, fg_col, bg_col, str);
}

int BlitPrintfCenter(int y, u32 fg_col, u32 bg_col, const char *format, ...)
{
	int x, len;
	va_list ap;
	char str[256];

	va_start(ap, format);
	len = vsnprintf(str, 256, format, ap);
	va_end(ap);

	x = 30 - len / 2;
	return BlitPrint(x, y, fg_col, bg_col, str);
}

int BlitPrintfRight(int y, u32 fg_col, u32 bg_col, const char *format, ...)
{
	int x, len;
	va_list ap;
	char str[256];

	va_start(ap, format);
	len = vsnprintf(str, 256, format, ap);
	va_end(ap);

	x = 60 - len;
	return BlitPrint(x, y, fg_col, bg_col, str);
}

int PutLine(int y, u32 fg_col, u32 bg_col, char ch)
{
	int i;
	char str[65];

	for(i = 0; i < 64; i++)
		str[i] = ch;

	str[64] = '\0';
	return BlitPrint(0, y, fg_col, bg_col, str);
}

void DrawLine(int y, u32 color)
{
	int i, n;
	u32 *vram32;
	u16 *vram16;
	u16 col16;

	y <<= 3;

	// draw
	if(draw_buf.pixelformat == PSP_DISPLAY_PIXEL_FORMAT_8888)
	{
		vram32 = (u32 *)draw_buf.vram + y * draw_buf.bufferwidth;

		for(i = 0; i < 8; i++)
		{
			for(n = 0; n < draw_buf.bufferwidth; n++)
			{
				*vram32++ = color;
			}
		}
	}
	else
	{
		col16 = convert_u32_to_u16(color, draw_buf.pixelformat);
		vram16 = (u16 *)draw_buf.vram + y * draw_buf.bufferwidth;

		for(i = 0; i < 8; i++)
		{
			for(n = 0; n < draw_buf.bufferwidth; n++)
			{
				*vram16++ = col16;
			}
		}
	}

	return;
}

