/*
	PluginManager v4
	Copyright (C) 2011, plum

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// header
#include "plugin_mgr.h"

// function_ptr
SceUID (*_sceKernelLoadModule)(const char *path, int flags, SceKernelLMOption *option);

// global
u32 patch_addr;
u32 restore[2];
int set_count;
BlackList *user_list;

// extern
extern char boot_path[256];
extern SceModuleContainer *mod_list;

// function
u32 FindImportAddr(SceModule2 *mod, const char *importlib, u32 nid)
{
	int i;
	u32 stub_top, stub_end;
	SceLibraryStubTable *stub;

	stub_top = (u32)mod->stub_top;
	stub_end = stub_top + mod->stub_size;

	while(stub_top < stub_end)
	{
		stub = (SceLibraryStubTable *)stub_top;

		if(strcmp(stub->libname, importlib) == 0)
		{
			for(i = 0; i < stub->stubcount; i++)
			{
				if(stub->nidtable[i] == nid)
					return (u32)(stub->stubtable + i * 8);
			}

			break;
		}

		stub_top += (stub->len * 4);
	}

	return 0;
}

u32 GetLoadModuleAddr(void)
{
	u32 nid;
	int devkit, version;

	devkit = sceKernelDevkitVersion();
	version = (((devkit >> 24) & 0xF) << 8) | (((devkit >> 16) & 0xF) << 4) | ((devkit >> 8) & 0xF);

	switch(version)
	{
		case 0x500:
		case 0x550:
			nid = 0xC5A281C5;
			break;

		case 0x620:
			nid = 0xE3CCC6EA;
			break;

		case 0x635:
		case 0x636:
		case 0x637:
		case 0x638:
		case 0x639:
			nid = 0xFFB9B760;
			break;

		case 0x660:
			nid = 0x939E4270;
			break;

		default:
			nid = 0;
			break;
	}

	if(!nid)
		return 0;

	return sctrlHENFindFunction("sceModuleManager", "ModuleMgrForKernel", nid);
}

void AdjustThreads(SceModuleContainer *top)
{
	int i, n, match;

	for(i = 0, match = 0; i < top->thid_count; i++)
	{
		if(top->thid[i] == 0)
		{
			for(n = i; n < top->thid_count; i++)
			{
				top->thid[n] = top->thid[n + 1];
			}

			match++;
		}
	}

	top->thid_count -= match;
	return;
}

void DeleteThreadId(SceModuleContainer *top)
{
	int i;
	SceKernelThreadRunStatus stat;

	for(i = 0; i < top->thid_count; i++)
	{
		if(sceKernelReferThreadRunStatus(top->thid[i], &stat) != 0)
		{
			top->thid[i] = 0;
		}
	}

	return;
}

void BackupThreadId(SceUID thid, SceModuleContainer *top)
{
	// delete
	DeleteThreadId(top);

	// adjust
	AdjustThreads(top);

	// backup
	top->thid[top->thid_count] = thid;
	top->thid_count++;
	return;
}

int sceKernelStartThreadPatched(SceUID thid, SceSize arglen, void *argp)
{
	int k1 = pspSdkSetK1(0);
	SceModuleContainer *top = mod_list;

	while(top != NULL)
	{
		if(top->number == arglen)
		{
			BackupThreadId(thid, top);
			break;
		}

		top = top->next;
	}

	pspSdkSetK1(k1);
	return sceKernelStartThread(thid, arglen, argp);
}

void PatchThreadStub(SceModule2 *mod, SceModuleContainer *top)
{
	// find sceKernelStartThread stub
	u32 orgaddr = FindImportAddr(mod, "ThreadManForKernel", 0xF475845D);

	if(orgaddr != 0)
	{
		// set flag
		top->patch_flag = 1;

		// patch stub
		MAKE_JUMP(orgaddr, sceKernelStartThreadPatched);
		_sw(0x24050000 | top->number, orgaddr + 4); // li $a1, top->number

		// clear
		ClearCaches();
	}

	return;
}

SceUID sceKernelLoadModulePatched(const char *path, int flags, SceKernelLMOption *option)
{
	int k1;
	char *ptr;
	SceUID modid;
	SceModule2 *mod;

	k1 = pspSdkSetK1(0);

	// restore
	_sw(restore[0], patch_addr + 0);
	_sw(restore[1], patch_addr + 4);
	ClearCaches();

	// call
	modid = _sceKernelLoadModule(path, flags, option);

	// find mod
	mod = sceKernelFindModuleByUID(modid);

	// set
	if(mod != NULL)
	{
		if(mod->attribute & PSP_MODULE_KERNEL)
		{
			if(strncmp(path, "ms0", 3) == 0 || strncmp(path, "ef0", 3) == 0)
			{
				if(strcmp(mod->modname, "RebootLoader") == 0)
				{
					strcpy(boot_path, path);
				}
				else
				{
					set_count++;
					SetContainer(path, mod, set_count);
				}
			}
		}
		else
		{
			ptr = strrchr(path, '/');
			SetBlackList(ptr + 1);
		}
	}

	// patch
	REDIRECT_FUNCTION(patch_addr, sceKernelLoadModulePatched);
	ClearCaches();

	pspSdkSetK1(k1);
	return modid;
}

void PatchModuleMgr(void)
{
	// init
	mod_list = NULL;
	set_count = 0;

	// get patch_addr
	patch_addr = GetLoadModuleAddr();

	if(!patch_addr)
		return;

	// backup
	restore[0] = _lw(patch_addr + 0);
	restore[1] = _lw(patch_addr + 4);

	// get function
	_sceKernelLoadModule = (void *)patch_addr;

	// patch
	REDIRECT_FUNCTION(patch_addr, sceKernelLoadModulePatched);
	ClearCaches();

	return;
}

