/*
	PluginManager v4
	Copyright (C) 2011, plum

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// header
#include "plugin_mgr.h"

// global
SceUID *start_threads, start_count;
SceUID *cur_threads, cur_count;

// extern
extern BlackList *user_list;
extern SceModuleContainer *mod_list;

// function
void *malloc_p(SceSize size)
{
	SceUID memid;
	void *ptr;

	memid = sceKernelAllocPartitionMemory(1, "block", 1, size + 4, NULL);

	if(memid < 0)
		return NULL;

	ptr = sceKernelGetBlockHeadAddr(memid);

	memcpy(ptr, &memid, 4);
	return ptr + 4;
}

int free_p(void *ptr)
{
	SceUID memid;

	if(!ptr)
		return -1;

	memcpy(&memid, ptr - 4, 4);
	return sceKernelFreePartitionMemory(memid);
}

int ReadLine(SceUID fd, char *buf, int size)
{
	char ch;
	int res, count;

	count = 0;
	size--;

	do
	{
		res = sceIoRead(fd, &ch, 1);

		if(res != 1)
			break;

		if(ch == ' ' || ch == '\r')
			continue;

		else if(ch == '\n')
			break;

		else
		{
			buf[count++] = ch;
		}
	}
	while(res == 1 && count < size);

	buf[count] = '\0';
	return count;
}

int SetContainer(const char *prxpath, SceModule2 *module, int count)
{
	char *pname;
	SceModuleContainer *ptr, *top, *backup;

	// alloc memory
	ptr = malloc_p(sizeof(SceModuleContainer));

	if(!ptr)
		return -1;

	// init
	ptr->next = NULL;
	ptr->prev = NULL;
	ptr->patch_flag = 0;
	ptr->stat = 1;
	ptr->thid_count = 0;
	memset(ptr->thid, 0, sizeof(SceUID) * 10);

	// set number
	ptr->number = count;

	// get name
	pname = strrchr(prxpath, '/');
	pname++;

	// get prxname
	ptr->prxname = malloc_p(strlen(pname) + 1);

	if(!ptr->prxname)
		return -1;

	strcpy(ptr->prxname, pname);

	// get mod
	memcpy(&ptr->mod, module, sizeof(SceModule2));

	// set next
	if(mod_list == NULL)
	{
		mod_list = ptr;
	}
	else
	{
		top = mod_list;

		while(top->next != NULL)
			top = top->next;

		top->next = ptr;

		// set prev
		backup = top->next;
		backup->prev = top;
	}

	return 0;
}

void AdjustContainer(SceModuleContainer *top, int flag)
{
	int check;
	SceModuleContainer *back = NULL;

	if(top != NULL)
	{
		check = (flag) ? top->thid_count : top->patch_flag;

		if(check == 0)
		{
			if(top->prev == NULL)
			{
				free_p(top->prxname);
				free_p(top);
				top = NULL;
				return;
			}

			back = top->prev;
			back->next = top->next;
			free_p(top->prxname);
			free_p(top);
		}
	}

	return;
}

int SetBlackList(const char *prxpath)
{
	BlackList *ptr, *top;

	if(!prxpath)
		return -1;

	// alloc
	ptr = malloc_p(sizeof(BlackList));

	if(!ptr)
		return -1;

	ptr->path = NULL;
	ptr->next = NULL;

	// get path
	ptr->path = malloc_p(strlen(prxpath) + 1);

	if(!ptr->path)
	{
		free_p(ptr);
		return -1;
	}

	strcpy(ptr->path, prxpath);

	// set next
	if(user_list == NULL)
	{
		user_list = ptr;
	}
	else
	{
		top = user_list;

		while(top->next != NULL)
			top = top->next;

		ptr->next = ptr;
	}

	return 0;
}

SceUID LoadStartModule(const char *path)
{
	SceUID res, modid;

	res = -1;
	modid = sceKernelLoadModule(path, 0, NULL);

	if(modid > 0)
	{
		res = sceKernelStartModule(modid, strlen(path)+1, (char *)path, NULL, NULL);
	}

	return res;
}

SceUID CreateStartThread(const char *name, SceKernelThreadEntry entry, int pri, int stack_size)
{
	SceUID thid = sceKernelCreateThread(name, entry, pri, stack_size, 0, NULL);

	if(thid >= 0)
	{
		sceKernelStartThread(thid, 0, NULL);
	}

	return thid;
}

int InitThreadList(void)
{
	start_threads = malloc_p(sizeof(SceUID) * 64);
	cur_threads = malloc_p(sizeof(SceUID) * 64);

	if(!start_threads || !cur_threads)
		return -1;

	start_count = 0;
	cur_count = 0;

	return sceKernelGetThreadmanIdList(SCE_KERNEL_TMID_Thread, start_threads, 64, &start_count);
}

void SuspendThreads(void)
{
	int i, n;
	SceUID cur_thid;
	SceKernelThreadInfo info;

	// init
	cur_count = 0;
	memset(cur_threads, 0, sizeof(SceUID) * 64);

	// get cur_thid
	cur_thid = sceKernelGetThreadId();

	// get threads
	if(sceKernelGetThreadmanIdList(SCE_KERNEL_TMID_Thread, cur_threads, 64, &cur_count) < 0)
		return;

	// suspend threads
	for(i = 0; i < cur_count; i++)
	{
		info.size = sizeof(SceKernelThreadInfo);

		if(sceKernelReferThreadStatus(cur_threads[i], &info) == 0)
		{
			if(info.status & PSP_THREAD_SUSPEND)
				continue;
		}

		for(n = 0; n < start_count; n++)
		{
			if(cur_threads[i] == start_threads[n])
			{
				cur_threads[i] = 0;
				n = start_count;
			}
		}

		if(cur_threads[i] == cur_thid)
			cur_threads[i] = 0;

		sceKernelSuspendThread(cur_threads[i]);
	}

	return;
}

void ResumeThreads(void)
{
	int i, n;
	SceModuleContainer *top;

	// resume_threads
	for(i = 0; i < cur_count; i++)
	{
		top = mod_list;

		while(top != NULL)
		{
			if(top->stat == 0)
			{
				for(n = 0; n < top->thid_count; n++)
				{
					if(cur_threads[i] == top->thid[n])
					{
						cur_threads[i] = 0;
						n = top->thid_count;
					}
				}
			}

			top = top->next;
		}

		sceKernelResumeThread(cur_threads[i]);
	}

	return;
}

